----------------------------------
--Traffic Generator				--
--Written by: Andreas Heps     --
----------------------------------

library ieee;
use ieee.std_logic_1164.all;  
use ieee.numeric_std.all;

entity TG is
port(
	ungated_clk	:in  std_logic;
	clk			:in  std_logic;
	reset		:in  std_logic;
	start_testing	: in std_logic;
	x_in		:in  std_logic_vector(7 downto 0);
	request		:in  std_logic;
	grant		:in  std_logic;
	SW_enable	:in  std_logic;
	SW_addr		:in  std_logic_vector(8 downto 0);
	global_time	:in std_logic_vector(31 downto 0);
	x_out		:out std_logic_vector(7 downto 0);
	RTS			:out std_logic;
	CTS			:out std_logic;
	tstamp_out	:out std_logic_vector(31 downto 0)
	);
end entity;


architecture behavioral of TG is
--------------------
--TYPE DECLERATION--
--------------------
type t_RecvState is (RST_BRAM,IDLE,SYNC,RECEIVE,CONF);
type t_WatchState is (IDLE, WATCHING);
type t_SendState is (IDLE,SYNC,SENDING);

----------------------
--SIGNAL DECLERATION--
---------------------- 
--Signals for storing last received package
signal packet_in					:std_logic_vector(63 downto 0);
signal packet_out					:std_logic_vector(63 downto 0);

--Counter signals for use with the statemachines
signal recv_counter					:INTEGER range -1 to 8;
signal send_counter					:integer range 0 to 8;

--Control signals for controling the response when receiving a packet
signal enable_timer					:std_logic;

--Signals for configuration of the traffic_generator
--signal config						:std_logic;
signal on_off						:std_logic;						--sending on or off
signal send_to						:std_logic_vector(3 downto 0);
signal delay						:std_logic_vector(11 downto 0); 	--the delay of the packet production
signal enable_processing_time		:std_logic;

--Signals for LFSR_8 entity
signal enable_LFSR_8				:std_logic;
signal random_8bit					:std_logic_vector(7 downto 0);

--Signals for BRAM entity
signal enable_BRAM		: std_logic;
signal wenable_BRAM		: std_logic;
signal addr_BRAM		: std_logic_vector(8 downto 0);
signal data_i_BRAM		: std_logic_vector(31 downto 0);

-----
signal send_package		: integer range 0 to 255;
signal send_counting	: integer range 0 to 12;

-------------------------
-- END OF DECLERATIONS --
-------------------------

begin
	
	LFSR_8:	entity work.LFSR_8(behavioral)
		port map(
		clk 	=> clk,
		reset 	=> reset,
		enable 	=> enable_LFSR_8,
		c_out 	=> random_8bit
		);
		
	BRAM: entity work.BRAM(behavioral)
		port map(
		ungated_clk	=> ungated_clk,
		clk			=> clk,	   
		enable		=> enable_BRAM,
		SW_enable	=> SW_enable,
		wenable		=> wenable_BRAM,
		addr		=> addr_BRAM,
		SW_addr		=> SW_addr,
		data_i		=> data_i_BRAM,
		data_o		=> tstamp_out
		);
	
	recv: process(clk,reset,on_off,send_to,global_time,packet_in,packet_out) is
	
	variable recv_state, nrecv_state :t_RecvState;
	variable send_state, nsend_state :t_SendState;
	variable watch_state, nwatch_state:t_WatchState;

	variable trigger 	: integer range 0 to 4096;
	
	variable config : std_logic; 
	
	variable packet_counter	: integer range 0 to 512;
	
	variable cnt : integer range 0 to 31;
    
	begin 
		
		
        if reset = '0' then
			config 		:= '0';
			on_off 		<= '0';
			send_to		<= (others=>'0');
			delay		<= (others=>'0');
			enable_processing_time <= '0';
				
			
			nrecv_state 	:= RST_BRAM;
			recv_state 		:= RST_BRAM;
			send_state		:= IDLE;
			nsend_state		:= IDLE;
			watch_state		:= IDLE;
			nwatch_state	:= IDLE;
			CTS 			<= '0';
			RTS 			<= '0';
			x_out 			<= (others=>'0');
			recv_counter 	<= -1;
			send_counter 	<= 0;			
			trigger 		:= 0;

			enable_LFSR_8 	<= '1';
			enable_timer 	<= '0';
			
			packet_in		<= (others=>'0');
			packet_out		<= (others=>'0');
			
			enable_BRAM <= '0';
			wenable_BRAM <= '0';
			data_i_BRAM <= (others => '0');
			cnt := 0;	 
			
			send_package <= 0;	
			send_counting <= 0;
		
        elsif clk'event and clk='1' then
		-----------
		--COUNTER--
		-----------
		if enable_timer = '0' then
			trigger := 0;
		else
			if trigger = 4096 then
				trigger := 0;
			else
				trigger := trigger + 1;
			end if;
		end if;
		
		-------------
		--FestState--
		-------------
		
		case watch_state is
			when IDLE =>
			if on_off = '1' and start_testing = '1' then 
				enable_LFSR_8 <= '0';
				enable_timer <= '1';
				if enable_processing_time = '1' then
					if trigger = to_integer(UNSIGNED(random_8bit)) then
						send_package <= send_package + 1;
						nwatch_state := WATCHING;
						enable_LFSR_8 <= '1'; --enable LFSR_8 pseudo_random counting
						enable_timer <= '0';
					end if;
				else
					if trigger = to_integer(UNSIGNED(delay)) then
						send_package <= send_package + 1;
						enable_timer <= '0';
						nwatch_state := WATCHING;
					end if;
				end if;
			end if;
			
			when WATCHING =>
				if send_counting = 10 then
					nwatch_state := IDLE;
					send_counting <= 0;
				else
					nwatch_state := WATCHING;
					send_counting <= send_counting + 1;
				end if;
		end case;
			
		
		-------------------------
		-- Output statemachine --
		-------------------------
		
		case send_state is
			when IDLE =>
			enable_BRAM <= '0';
			wenable_BRAM <= '0';
			RTS <= '0';
				
			if send_package /= 0 then
				nsend_state := SYNC;
				packet_out <= send_to & "0000" & x"000000" & global_time;
			else
				nsend_state := IDLE;
			end if;
									
			when SYNC =>				 
			--wait for synchronization with the "send_to" router.
			RTS <= '1';
			if grant = '1' then
				nsend_state := SENDING;
				send_counter <= send_counter + 1;
				packet_out <= packet_out(55 downto 0) & packet_out(63 downto 56);
			end if;
			
			when SENDING =>
			--send the generated packet out on the network
			--x_out <= "00000000";
			if send_counter = 7 then
				if grant = '1' then
					nsend_state := IDLE;
					send_counter <= 0;
					RTS <= '0';
					
				end if;
			elsif send_counter = 6 then
				send_counter <= send_counter + 1;
				send_package <= send_package - 1;
				packet_out <= packet_out(55 downto 0) & packet_out(63 downto 56);
			else
				send_counter <= send_counter + 1;
				packet_out <= packet_out(55 downto 0) & packet_out(63 downto 56);
			end if;
			
			when OTHERS =>
			null;
			end case;
			
		------------------------
		-- Input statemachine --
		------------------------
		
		case recv_state is
			
			when RST_BRAM =>
			enable_BRAM <= '1';
			wenable_BRAM <= '1';
			addr_BRAM <= std_logic_vector(to_signed(cnt, 9));
			if cnt = 31 then
				nrecv_state := IDLE;
			else
				nrecv_state := RST_BRAM;
				cnt := cnt + 1;
			end if;			
		
			when IDLE =>
			if request = '1' then
				nrecv_state := SYNC;
				CTS <= '1';
				recv_counter <= recv_counter + 1;
			else
				nrecv_state := IDLE;
				CTS <= '0';
			end if;

			when SYNC =>
				nrecv_state := RECEIVE;
			
			when RECEIVE =>
			CTS <= '1';
			
			if recv_counter = 8 then
				if config = '1' then
					nrecv_state := CONF;
					config := '0';
				else				
					nrecv_state := IDLE;
					if(packet_counter = 512) then
						packet_counter := 0;
					end if;
						
					--Writing tstamp to BRAM
					addr_BRAM <= std_logic_vector(to_unsigned(packet_counter, 9));
					enable_BRAM <= '1';
					wenable_BRAM <= '1';
					data_i_BRAM <= std_logic_vector(to_unsigned((to_integer(UNSIGNED(global_time))-to_integer(UNSIGNED(packet_in(31 downto 0)))),32));
					packet_counter := packet_counter + 1;

				end if;
				CTS <= '0';
				recv_counter <= -1;
			else
				if recv_counter = 7 then
					CTS <= '0';
				elsif recv_counter = 0 and x_in(3) = '1' then
					config := '1';
				end if;
				
				recv_counter <= recv_counter + 1;
				packet_in <= packet_in(55 downto 0) & x_in;
			end if;
			
			when CONF=>	
			on_off	 				<= packet_in(58);
			enable_processing_time 	<= packet_in(57);
			send_to					<= packet_in(55 downto 52);
			delay					<= packet_in(51 downto 40);
			trigger		:= 0;
			nrecv_state := IDLE;
			
		end case;				  
		--Driving the next state signal
		recv_state := nrecv_state;
		send_state := nsend_state;
		watch_state := nwatch_state;
		--Driving the x_out signal
		x_out <= packet_out(63 downto 56);
			
		end if;
	end process;
		
end architecture;